@@ -10,18 +10,11 @@ upgrade = VersionUpgrade21to22.VersionUpgrade21to22()
 
 def getMetaData():
     return {
-        "plugin": {
-            "name": catalog.i18nc("@label", "Version Upgrade 2.1 to 2.2"),
-            "author": "Ultimaker",
-            "version": "1.0",
-            "description": catalog.i18nc("@info:whatsthis", "Upgrades configurations from Cura 2.1 to Cura 2.2."),
-            "api": 3
-        },
         "version_upgrade": {
-            # From                     To                 Upgrade function
-            ("profile", 1):          ("quality", 2,       upgrade.upgradeProfile),
-            ("machine_instance", 1): ("machine_stack", 2, upgrade.upgradeMachineInstance),
-            ("preferences", 2):      ("preferences", 3,   upgrade.upgradePreferences)
+            # From                         To                         Upgrade function
+            ("profile", 1000000):          ("quality", 2000000,       upgrade.upgradeProfile),
+            ("machine_instance", 1000000): ("machine_stack", 2000000, upgrade.upgradeMachineInstance),
+            ("preferences", 2000000):      ("preferences", 3000000,   upgrade.upgradePreferences)
         },
         "sources": {
             "profile": {
